#define _XOPEN_SOURCE 500
#include "read_binvox.h"
#include <getopt.h>
#include <math.h>
#include <mpi.h>
#include <netcdf.h>
#include <netcdf_par.h>
#include <omp.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef LIKWID_PERFMON
#include <likwid-marker.h>
#else
#define LIKWID_MARKER_INIT
#define LIKWID_MARKER_THREADINIT
#define LIKWID_MARKER_SWITCH
#define LIKWID_MARKER_REGISTER(regionTag)
#define LIKWID_MARKER_START(regionTag)
#define LIKWID_MARKER_STOP(regionTag)
#define LIKWID_MARKER_CLOSE
#define LIKWID_MARKER_GET(regionTag, nevents, events, time, count)
#endif

#ifdef OSTREAM_LOOP_INDIRECT
#define P_COORDS
#endif

#ifdef OCOLLIDE_LOOP_INDIRECT
#define P_COORDS
#endif

/* Compile-time selection: use 32bit or 64bit fundamental data types */
#ifdef WITH_32BIT
typedef float real_t;
typedef int32_t int_t;
#define INT_FMT "d"
#define REAL_FMT "f"
#else
typedef double real_t;
typedef int64_t int_t;
#define INT_FMT "ld"
#define REAL_FMT "lf"
#endif

#define ERR                                                                    \
  {                                                                            \
    if (err != NC_NOERR)                                                       \
      printf("Error at line=%d: %s\n", __LINE__, nc_strerror(err));            \
  }

/* Simulation parameters. Default values can be overridden by options(). */
int_t g_height = 40, /* Vertical coordinates (index i)   */
  g_width = 80,      /* Horizontal coordinates (index j) */
  g_depth = 40,      /* Depth coordinates (index k) */
  snap_freq = 100,   /* Number of iterations between saved states */
  max_iter = 40001,  /* Total number of iterations */
  threads = 0,       /* Threads per CPU node*/
  cart_dims_opt[2] = { 0, 0 };
bool quiet = false; /* Suppresses stdio and file I/O when true */
bool debug = false;
char binvox_path[256], perf_file[256] = "";
bool load_binvox_data = false, store_binvox = false;

typedef enum
{
  POISSEUILLE,
  PIPE,
} domain_type_t;

struct domain_settings
{
  domain_type_t type;
  // How much of the domain should be fluid nodes. From 0-100
  int node_usage;
};

struct domain_settings domain_settings = { .type = POISSEUILLE,
                                           .node_usage = 100 };

#define DIMS 3
#define MAX_CART_DIMS 2
#define D_COUNT 27

int world_size, world_rank, cart_rank;
MPI_Comm comm_cart; // Communicator for the cartesian topology
int cart_coords[DIMS], cart_dims[DIMS];
MPI_Datatype
  // A lattice point
  lattice_pt,
  // Row/column types
  row, column, z_row, row_INT,
  // Domain types for mpi files
  subdomain, domain,
  // Normal row/column plane types. 64 bit types
  row_col_plane, row_dep_plane, col_dep_plane,
  // plane types with INT size (4 bytes) for MAP.
  row_dep_plane_INT, row_col_plane_INT,
  // Indexed plane types
  mpi_t_plane_S, mpi_t_plane_N, mpi_t_plane_W, mpi_t_plane_E,
  // Indexed plane types for ghost cells
  mpi_t_plane_S_ghost, mpi_t_plane_N_ghost, mpi_t_plane_W_ghost,
  mpi_t_plane_E_ghost,
  // Map domain subarray
  map_domain;

int neighbors[DIMS][2];
int_t local_size[DIMS], local_ghost_size[DIMS], height, width, depth,
  l_tot_size;

int ncid, dimid[3], var_map, varid, var_vz, var_vy, var_vx, var_rank;

// Number of fluid nodes in the domain. Used for indirect indexing
int_t fluid_node_count = 0;

/* Map of points to rank */
int* rank;

/* Indirect index array. */
int_t* index_arr;
struct coords
{
  int_t x;
  int_t y;
  int_t z;
};

struct coords* p_coords;

/* Macros for N-S-E-W neighbors of the current process.
 * Y = dim 0, X = dim 1
 * Positive direction down, right
 */
#define NB_N neighbors[0][0]
#define NB_S neighbors[0][1]
#define NB_W neighbors[1][0]
#define NB_E neighbors[1][1]

/* Lattice geometry and physical data */
real_t
  /* Density distributions for present and next iteration */
  *density[2] = { NULL, NULL },

  /* 3D velocity vectors derived from the present density distribution */
  *velocity = NULL,

  /* Magnitude of velocity vectors, for visualization purposes */
    *abs_velocity = NULL,

    *v_x,
    *v_y,
    *v_z,

/* Sine and cosine values of the lattice directions */
    c[D_COUNT][DIMS] = {
      { 0, 0, 0 }, // Rest
      // Z = 0
      { 0, 1, 0 },   // Right
      { 1, 1, 0 },  // Right/up
      { 1, 0, 0 },  // Up
      { 1, -1, 0 }, // Left/up
      { 0, -1, 0 },  // Left
      { -1, -1, 0 },  // Left/down
      { -1, 0, 0 },   // Down
      { -1, 1, 0 },   // Right/down
      // Z = -1
      { 0, 0, -1 },   // back
      { 0, 1, -1 },   // back right
      { 1, 1, -1 },  // back right up
      { 1, 0, -1 },  // back up
      { 1, -1, -1 }, // back left up
      { 0, -1, -1 },  // back left
      { -1, -1, -1 },  // back down left
      { -1, 0, -1 },   // back down
      { -1, 1, -1 },   // back down right
      // Z = +1
      { 0, 0, 1 },   // front
      { 0, 1, 1 },   // front right
      { 1, 1, 1 },  // front right up
      { 1, 0, 1 },  // front up
      { 1, -1, 1 }, // front left up
      { 0, -1, 1 },  // front left
      { -1, -1, 1 },  // front down left
      { -1, 0, 1 },   // front down
      { -1, 1, 1 },   // front down right
    },

  /* Weight coefficient of relaxation operator */
  lambda = -1.0,

  /* Artificial external force of inflow */
  force[DIMS] = { 0.0, 1e-1, 0.0 },

  /* Radius of the bluff obstruction.
   * This value should be positive, its default value is computed by
   * init_domain() unless it is overridden by a command line argument.
   */
  radius = 0.0,
  pipe_radius = 0.0,

  /*
   * Weights for all the lattice points.
   * Taken from 10.1016/j.camwa.2015.01.010
   * A D2Q27 multiple-relaxation-time lattice Boltzmann method
   */
  w[D_COUNT] = {
      8.0/27.0,   // Rest
      2.0/27.0,   // Right
      1.0/54.0,   // Right/up
      2.0/27.0,   // Up
      1.0/54.0,   // Left/up
      2.0/27.0,   // Left
      1.0/54.0,   // Left/down
      2.0/27.0,   // Down
      1.0/54.0,   // Right/down
      // Z - 1
      2.0/27.0,   // back
      1.0/54.0,   // back right
      1.0/216.0,  // back right up
      1.0/54.0,   // back up
      1.0/216.0,  // back left up
      1.0/54.0,   // back left
      1.0/216.0,  // back down left
      1.0/54.0,   // back down
      1.0/216.0,  // back down right
      // Z + 1
      2.0/27.0,   // front
      1.0/54.0,   // front right
      1.0/216.0,  // front right up
      1.0/54.0,   // front up
      1.0/216.0,  // front left up
      1.0/54.0,   // front left
      1.0/216.0,  // front down left
      1.0/54.0,   // front down
      1.0/216.0,  // front down right
  },

  dt = 1.0, dx = 1.0, c_s;

/* y and x coordinate shifts to neighboring lattice nodes */

/*
 * For D3Q27, the rest is at 0, 1 is right, 2 right/up ... until 8 = right down
 * We continue the count at z-1 and z+1, so 9 is "right behind" (z-1 from
 * center) Then 18 is right in front, so z+1 from center
 */

int_t offsets[D_COUNT][DIMS] = {
  // Rest
  { 0, 0, 0 }, // Rest
  // Z = 0
  { 0, 1, 0 },   // Right
  { -1, 1, 0 },  // Right/up
  { -1, 0, 0 },  // Up
  { -1, -1, 0 }, // Left/up
  { 0, -1, 0 },  // Left
  { 1, -1, 0 },  // Left/down
  { 1, 0, 0 },   // Down
  { 1, 1, 0 },   // Right/down
  // Z = -1
  { 0, 0, -1 },   // back
  { 0, 1, -1 },   // back right
  { -1, 1, -1 },  // back right up
  { -1, 0, -1 },  // back up
  { -1, -1, -1 }, // back left up
  { 0, -1, -1 },  // back left
  { 1, -1, -1 },  // back down left
  { 1, 0, -1 },   // back down
  { 1, 1, -1 },   // back down right
  // Z = +1
  { 0, 0, 1 },   // front
  { 0, 1, 1 },   // front right
  { -1, 1, 1 },  // front right up
  { -1, 0, 1 },  // front up
  { -1, -1, 1 }, // front left up
  { 0, -1, 1 },  // front left
  { 1, -1, 1 },  // front down left
  { 1, 0, 1 },   // front down
  { 1, 1, 1 },   // front down right
};

#define OFFSET(d, axis) offsets[d][axis]

/* Map of the simulated domain, this defines the problem geometry.
 * initialize_domain() sets up a set of SOLID/FLUID points, and
 * then calculates which of the SOLID points are actually WALL,
 * on account of having at least one FLUID neighbor.
 * SOLID points are omitted by the compute kernels, as they have no
 * direct impact on fluid flow.
 */
typedef enum
{
  FLUID,
  WALL,
  SOLID
} node_type_t;
node_type_t* map = NULL;
node_type_t* map_final = NULL;
#define MAP(i, j, k)                                                           \
  map[(width * (height + 2) * (k + 1)) + (width * (i + 1)) + (j)]
#define MAP_I(i, j, k) map_final[k * height * width + i * width + j]

/* Compile-time configurable memory layout of the 6 densities associated
 * with a lattice point.
 * If PACK_DIRECTION_MAJOR is defined, densities in direction 0 for all
 * (i,j) are stored first, densities in direction 1 are stored next, etc.
 * Without PACK_DIRECTION_MAJOR, all 6 densities for each lattice point
 * are stored consecutively.
 */
#ifdef PACK_DIRECTION_MAJOR

#define D_now(i, j, d) density[0][(d)*width * height + (i)*width + (j)]
#define D_nxt(i, j, d) density[1][(d)*width * height + (i)*width + (j)]
#define V(i, j, x) velocity[2 * ((i)*width + (j)) + (x)]
#define V_abs(i, j) abs_velocity[(i)*width + (j)]

#else

// Get the index of the fluid node at position (i, j, k). Might be -1
#define P_index(i, j, k)                                                       \
  index_arr[(((width) * (height + 2)) * (k + 1)) + (width) * (i + 1) + (j)]

// Current density deistribution at (i, j, k) of lattice point d
#define D_now(i, j, k, d) density[0][D_COUNT * P_index(i, j, k) + (d)]

// Next density deistribution at (i, j, k) of lattice point d
#define D_nxt(i, j, k, d) density[1][D_COUNT * P_index(i, j, k) + (d)]

// Velocity at (i, j, k) in direction of dimension x
#define V(i, j, k, x)                                                          \
  velocity[DIMS * ((width * height * (k)) + (i)*width + (j)) + (x)]

// Magintude of velocity at point (i, j, k)
#define V_abs(i, j, k) abs_velocity[(k * width * height) + (i)*width + (j)]

#endif

int bounce_map[D_COUNT] = { 0,  5,  6,  7,  8,  1,  2,  3,  4,
                            18, 23, 24, 25, 26, 19, 20, 21, 22,
                            9,  14, 15, 16, 17, 10, 11, 12, 13 };

int
bounce(int_t d)
{
  return bounce_map[d];
}

/* Functions used in the time integration loop */
/* Configure problem geometry: periodic horizontal channel with a
 * bluff (circular) obstruction in the middle of the flow.
 */
void
initialize_domain(void);

void
initialize_premade_domain(Binvox_content* binvox_content);

void
find_bounding_box(Binvox_content* binvox_content,
                  int_t* depth,
                  int_t* width,
                  int_t* height);

/* Relax the densities at all lattice sites toward equilibrium. */
void
collide(void);
/* Propagate densities from each lattice site to its neighbors */
void
stream(void);
/* Save simulation state in gnuplot binary matrix format.
 * File names are numbered by iter / snap_freq, and stored in 'data/'
 */
void
store(int_t iter);

void
mpi_store(int_t iter);

/* Border exchange of ghost cells, required so all processes have the result
 * from the neighbor, in order to run the collide step properly.
 */
void
border_exchange(void);

/* Command line option parsing */
void
options(int argc, char** argv);

/*
 * Write performance characteristics to a csv file.
 */
void
mpi_write_perf(double collide_total,
               double stream_total,
               double exchange_total,
               int iterations);

void
create_cart(MPI_Comm communicator,
            int ndims,
            int* cart_coords,
            int* cart_dims /*, MPI_Comm subcomms[ndims]*/)
{
  // Loosely based off Dalcin et al 2018 (Fast parallel multidimensional FFT
  // using advanced MPI)
  //
  // This has been heavily modified. Might not be relevant anymore as we don't
  // use subcomms or any other fancy stuff as Dalcin et al.

  int process_count, periods[ndims], remdims[ndims];
  for (int i = 0; i < ndims; i++) {
    cart_dims[i] = periods[i] = remdims[i] = 0;
    cart_dims[i] = cart_dims_opt[i];
  }

  //  Wrap around in x direction
  /*periods[1] = 1;*/

  MPI_Comm_size(communicator, &process_count);
  MPI_Dims_create(process_count, ndims, cart_dims);
  MPI_Cart_create(communicator, ndims, cart_dims, periods, 1, &comm_cart);
  MPI_Comm_rank(comm_cart, &cart_rank);
  MPI_Cart_coords(comm_cart, cart_rank, ndims, cart_coords);

  if (world_rank == 0)
    printf("Configuring cartesian MPI topology:\n"
           "\ttotal process count (world size): %d\n"
           "\tdims: %d\n"
           "\tcart coords: %d, %d\n"
           "\tcart dimensions: %d, %d\n"
           "\tperiods: %d, %d\n"
           "\n",
           world_size,
           ndims,
           cart_coords[0],
           cart_coords[1],
           cart_dims[0],
           cart_dims[1],
           periods[0],
           periods[1]);

  // Find the neighbor ranks in all directions of the current rank
  for (int i = 0; i < ndims; i++) {
    MPI_Cart_shift(comm_cart, i, 1, &neighbors[i][0], &neighbors[i][1]);
  }
  printf("rank %d: (%d, %d): N: %d S: %d E: %d W: %d \n",
         cart_rank,
         cart_coords[0],
         cart_coords[1],
         NB_N,
         NB_S,
         NB_E,
         NB_W);
}

int_t
map_direct_to_indirect(int_t* indices,
                       int_t* local_indexes,
                       int* displacements,
                       MPI_Datatype* old_type,
                       MPI_Datatype* new_type,
                       int_t count)
{

  // Initialize border index array
  int_t node_count = 0;
  for (int_t i = 0; i < l_tot_size; i++) {
    local_indexes[i] = -1;
  }

  // Copy border indices to local array
  MPI_Gather(
    indices, 1, *old_type, local_indexes, count, MPI_INT64_T, 0, MPI_COMM_SELF);

  // Add fluid node indices to displacement array
  for (int_t i = 0; i < count; i++) {
    if (local_indexes[i] >= 0) {
      displacements[node_count++] = local_indexes[i];
    }
  }

  // Create indexed block for the border
  MPI_Type_create_indexed_block(
    node_count, 1, displacements, lattice_pt, new_type);

  MPI_Type_commit(new_type);

  return node_count;
}

void
create_types_indirect()
{
  if (world_rank == 0)
    printf("Configuring types for indirect addressing\n");

  /*
   * Depth, column and row types. In order of major: z -> y -> x
   */

  int_t* local_indexes = malloc(l_tot_size * sizeof(int_t));
  int* displacements = malloc(fluid_node_count * sizeof(int));

  int_t node_count_res;

  node_count_res =
    map_direct_to_indirect(&P_index(-1, 0, -1),
                           local_indexes,
                           displacements,
                           &row_dep_plane,
                           &mpi_t_plane_N_ghost,
                           local_ghost_size[1] * local_ghost_size[2]);

  /*printf("[%d] N ghost plane: \t%ld\n", cart_rank, node_count_res);*/

  node_count_res =
    map_direct_to_indirect(&P_index(0, 0, -1),
                           local_indexes,
                           displacements,
                           &row_dep_plane,
                           &mpi_t_plane_N,
                           local_ghost_size[1] * local_ghost_size[2]);
  /*printf("[%d] N       plane: \t%ld\n", cart_rank, node_count_res);*/

  node_count_res =
    map_direct_to_indirect(&P_index(height - 1, 0, -1),
                           local_indexes,
                           displacements,
                           &row_dep_plane,
                           &mpi_t_plane_S,
                           local_ghost_size[1] * local_ghost_size[2]);

  /*printf("[%d] S       plane: \t%ld\n", cart_rank, node_count_res);*/

  node_count_res =
    map_direct_to_indirect(&P_index(height, 0, -1),
                           local_indexes,
                           displacements,
                           &row_dep_plane,
                           &mpi_t_plane_S_ghost,
                           local_ghost_size[1] * local_ghost_size[2]);

  /*printf("[%d] S ghost plane: \t%ld\n", cart_rank, node_count_res);*/

  node_count_res =
    map_direct_to_indirect(&P_index(-1, 0, 0),
                           local_indexes,
                           displacements,
                           &row_col_plane,
                           &mpi_t_plane_W,
                           local_ghost_size[0] * local_ghost_size[1]);
  printf("[%d] W       plane: \t%ld\n", cart_rank, node_count_res);

  node_count_res =
    map_direct_to_indirect(&P_index(-1, 0, -1),
                           local_indexes,
                           displacements,
                           &row_col_plane,
                           &mpi_t_plane_W_ghost,
                           local_ghost_size[0] * local_ghost_size[1]);
  printf("[%d] W ghost plane: \t%ld\n", cart_rank, node_count_res);

  node_count_res =
    map_direct_to_indirect(&P_index(-1, 0, depth - 1),
                           local_indexes,
                           displacements,
                           &row_col_plane,
                           &mpi_t_plane_E,
                           local_ghost_size[0] * local_ghost_size[1]);
  printf("[%d] E       plane: \t%ld\n", cart_rank, node_count_res);

  node_count_res =
    map_direct_to_indirect(&P_index(-1, 0, depth),
                           local_indexes,
                           displacements,
                           &row_col_plane,
                           &mpi_t_plane_E_ghost,
                           local_ghost_size[0] * local_ghost_size[1]);
  printf("[%d] E ghost plane: \t%ld\n", cart_rank, node_count_res);

  free(local_indexes);
  free(displacements);
}

void
create_types(int ndims,
             int* local_size,
             int_t* local_ghost_size,
             int* local_origin,
             int* global_size)
{
  if (world_rank == 0)
    printf("Configuring types:\n"
           "\tdims: %d\n"
           "\tglobal size: %d, %d\n"
           "\tlocal size: %d, %d\n"
           "\tlocal origin: %d, %d\n"
           "\tlocal ghost size: %ld, %ld\n"
           "\n",
           ndims,
           global_size[0],
           global_size[1],
           local_size[0],
           local_size[1],
           local_origin[0],
           local_origin[1],
           local_ghost_size[0],
           local_ghost_size[1]);

  MPI_Type_contiguous(D_COUNT, MPI_DOUBLE, &lattice_pt);
  MPI_Type_commit(&lattice_pt);

  /*
   * Depth, column and row types. In
   * order of major: z -> y -> x
   */
  MPI_Type_vector(local_ghost_size[2],
                  1,
                  local_ghost_size[0] * local_ghost_size[1],
                  MPI_INT64_T,
                  &z_row);
  MPI_Type_vector(
    local_ghost_size[0], 1, local_ghost_size[1], MPI_INT64_T, &column);
  MPI_Type_vector(
    1, local_ghost_size[1], local_ghost_size[1], MPI_INT64_T, &row);
  MPI_Type_vector(
    1, local_ghost_size[1], local_ghost_size[1], MPI_INT, &row_INT);

  MPI_Type_commit(&column);
  MPI_Type_commit(&row);
  MPI_Type_commit(&z_row);
  MPI_Type_commit(&row_INT);

  // x-y plane type
  MPI_Type_vector(
    1, local_ghost_size[0], local_ghost_size[0], row, &row_col_plane);
  MPI_Type_vector(
    1, local_ghost_size[0], local_ghost_size[0], row_INT, &row_col_plane_INT);

  // x-z plane type
  MPI_Type_vector(
    local_ghost_size[2], 1, local_ghost_size[0], row, &row_dep_plane);
  MPI_Type_vector(
    local_ghost_size[2], 1, local_ghost_size[0], row_INT, &row_dep_plane_INT);

  // y-z plane type
  MPI_Type_vector(
    local_ghost_size[2], 1, local_ghost_size[1], column, &col_dep_plane);

  MPI_Type_commit(&row_col_plane);
  MPI_Type_commit(&row_dep_plane);
  MPI_Type_commit(&row_dep_plane_INT);
  MPI_Type_commit(&row_col_plane_INT);
  MPI_Type_commit(&col_dep_plane);

  // Map domain subarray
  MPI_Type_create_subarray(DIMS,
                           global_size,
                           local_size,
                           local_origin,
                           MPI_ORDER_C,
                           MPI_INT,
                           &map_domain);
  MPI_Type_commit(&map_domain);
}

/* Main program stages:
 *  - argument parsing
 *  - memory allocation
 *  - initialization
 *  - time integration loop
 *  - deallocation
 */
int
main(int argc, char** argv)
{

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &world_size);
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);

  if (world_rank == 0)
    options(argc, argv);

  // Broadcasting arguments
  MPI_Bcast(&g_width, 1, MPI_INT64_T, 0, MPI_COMM_WORLD);
  MPI_Bcast(&g_height, 1, MPI_INT64_T, 0, MPI_COMM_WORLD);
  MPI_Bcast(&g_depth, 1, MPI_INT64_T, 0, MPI_COMM_WORLD);
  MPI_Bcast(&radius, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
  MPI_Bcast(&threads, 1, MPI_INT64_T, 0, MPI_COMM_WORLD);
  MPI_Bcast(&force, DIMS, MPI_DOUBLE, 0, MPI_COMM_WORLD);
  MPI_Bcast(&snap_freq, 1, MPI_INT64_T, 0, MPI_COMM_WORLD);
  MPI_Bcast(&debug, 1, MPI_C_BOOL, 0, MPI_COMM_WORLD);
  MPI_Bcast(&max_iter, 1, MPI_INT64_T, 0, MPI_COMM_WORLD);
  MPI_Bcast(&load_binvox_data, 1, MPI_C_BOOL, 0, MPI_COMM_WORLD);
  MPI_Bcast(&domain_settings.type, 1, MPI_INT, 0, MPI_COMM_WORLD);
  MPI_Bcast(&domain_settings.node_usage, 1, MPI_INT, 0, MPI_COMM_WORLD);
  MPI_Bcast(&cart_dims_opt, 2, MPI_LONG, 0, MPI_COMM_WORLD);

  int perf_file_len = strlen(perf_file);

  MPI_Bcast(&perf_file_len, 1, MPI_INT, 0, MPI_COMM_WORLD);
  MPI_Bcast(&perf_file, perf_file_len, MPI_CHAR, 0, MPI_COMM_WORLD);

  // Create cartesian topology
  create_cart(MPI_COMM_WORLD, 2, cart_coords, cart_dims);

  // Debugging
  if (debug) {
    volatile int i = 0;
    char hostname[256];
    gethostname(hostname, sizeof(hostname));
    printf("PID %d on %s ready for attach\n", getpid(), hostname);
    fflush(stdout);
    while (0 == i)
      sleep(5);
  }

  if (threads == 0)
    omp_set_num_threads(omp_get_max_threads());
  else
    omp_set_num_threads(threads);
  if (world_rank == 0)
    printf("OpenMP thread count (per "
           "process): %d\n\n",
           omp_get_max_threads());

  if (world_rank == 0) {
    printf("\n"
           "Domain and program options:\n"
           "\tLattice point count: %d\n"
           "\tGrid size: x: %ld  y: %ld\n"
           "\tIterations: %ld\n",
           D_COUNT,
           width,
           height,
           max_iter);
  }

  c_s = (1.0 / sqrt(3.0)) * (dx / dt);

  Binvox_content* binvox_data = NULL;

  if (load_binvox_data) {
    binvox_data = malloc(sizeof(Binvox_content));

    if (world_rank == 0) {
      if (read_binvox(binvox_path, binvox_data) != 0) {
        printf("Error reading binvox "
               "data. Exiting\n");
        free(binvox_data);
        exit(1);
      }
      find_bounding_box(binvox_data, &g_depth, &g_width, &g_height);

      // Make sure that all ranks have the same local dimensions
      g_height += g_height % cart_dims[0];
      g_depth += g_depth % cart_dims[1];
    }

    MPI_Bcast(&binvox_data->height, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&binvox_data->width, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&binvox_data->depth, 1, MPI_INT, 0, MPI_COMM_WORLD);

    MPI_Bcast(&g_depth, 1, MPI_INT64_T, 0, MPI_COMM_WORLD);
    MPI_Bcast(&g_width, 1, MPI_INT64_T, 0, MPI_COMM_WORLD);
    MPI_Bcast(&g_height, 1, MPI_INT64_T, 0, MPI_COMM_WORLD);

    int_t size = binvox_data->height * binvox_data->width * binvox_data->depth;
    if (world_rank != 0) {
      // Allocate space in all ranks for
      // the binvox data
      binvox_data->data = malloc(size);
    }
    MPI_Barrier(MPI_COMM_WORLD);
    MPI_Bcast(binvox_data->data, size, MPI_CHAR, 0, MPI_COMM_WORLD);
  }

  int local_size[DIMS] = { g_height / cart_dims[0],
                           g_width,
                           g_depth / cart_dims[1] };
  local_ghost_size[0] = local_size[0] + 2;
  local_ghost_size[1] = local_size[1];
  local_ghost_size[2] = local_size[2] + 2;
  int global_coords[DIMS] = { cart_coords[0] * local_size[0],
                              0,
                              cart_coords[1] * local_size[2] };
  int global_size[DIMS] = { g_height, g_width, g_depth };
  create_types(2, local_size, local_ghost_size, global_coords, global_size);
  height = local_size[0];
  width = local_size[1];
  depth = local_size[2];

  l_tot_size = local_ghost_size[0] * local_ghost_size[1] * local_ghost_size[2];
  index_arr = malloc(l_tot_size * sizeof(int_t));
#ifdef P_COORDS
  p_coords = malloc(l_tot_size * sizeof(struct coords));
#endif

  map = malloc(l_tot_size * sizeof(node_type_t));
  map_final = malloc(width * height * depth * sizeof(node_type_t));
  rank = malloc(width * height * depth * sizeof(int));

  /* Set up the simulated geometry */
  if (load_binvox_data) {
    initialize_premade_domain(binvox_data);
    free(binvox_data->data);
    free(binvox_data);
  } else
    initialize_domain();

  int_t fluid_count_sum = 0, fluid_count_local = 0;
  for (int_t k = 0; k < depth; k++)
    for (int_t i = 0; i < height; i++)
      for (int_t j = 0; j < width; j++) {
        if (P_index(i, j, k) >= 0)
          fluid_count_local++;
      }
  MPI_Reduce(&fluid_count_local,
             &fluid_count_sum,
             1,
             MPI_INT64_T,
             MPI_SUM,
             0,
             comm_cart);
  if (world_rank == 0) {
    float percent =
      ((double)fluid_node_count / (double)(width * height * depth)) * 100;
    float percent_g =
      ((double)fluid_count_sum / (double)(g_width * g_height * g_depth)) * 100;
    printf("\nDOMAIN PAREMETERS (rank 0)\n"
           "\tDensity node count: %ld\n"
           "\t |-without borders: %ld\n"
           "\tTotal node count:   %ld\n"
           "\tDomain use:         %f%%\n"
           "\tTOTAL Domain use    %f%%\n"
           "\n",
           fluid_node_count,
           fluid_count_local,
           width * height * depth,
           percent,
           percent_g);
  }

  create_types_indirect();

  /* Make all required dynamic memory
   * allocations, with space for ghost
   * cells
   */
  density[0] = malloc(D_COUNT * fluid_node_count * sizeof(real_t));
  density[1] = malloc(D_COUNT * fluid_node_count * sizeof(real_t));
  velocity = malloc(DIMS * width * height * depth * sizeof(real_t));
  abs_velocity = malloc(width * height * depth * sizeof(real_t));

#ifdef P_COORDS
  p_coords = realloc(p_coords, fluid_node_count * sizeof(struct coords));
#endif

  v_x = malloc(width * height * depth * sizeof(real_t));
  v_y = malloc(width * height * depth * sizeof(real_t));
  v_z = malloc(width * height * depth * sizeof(real_t));

  /* Initialize density distribution to
   * equilibrium, i.e. assume a density
   * of 1 per lattice point, and
   * distribute it evenly in 6
   * directions.
   */
  for (int_t k = -1; k < depth + 1; k++)
    for (int_t i = -1; i < height + 1; i++)
      for (int_t j = 0; j < width; j++) {

        int_t index = P_index(i, j, k);
        if (index >= 0)
          for (int_t d = 0; d < D_COUNT; d++)
            D_now(i, j, k, d) = D_nxt(i, j, k, d) = w[d];

        if (k >= 0 && k < depth && i >= 0 && i < height)
          MAP_I(i, j, k) = MAP(i, j, k);
      }

  double itertime, start_time, stream_time, collide_time, exchange_time;
  double stream_total = 0, collide_total = 0, exchange_total = 0;
  double last_snap = 0;

  // --------------- PROFILING START
  // ---------------------
  LIKWID_MARKER_INIT;
  MPI_Pcontrol(1);

  MPI_Barrier(comm_cart);

  /* Time integration loop */
  for (int_t iter = 0; iter < max_iter; iter++) {
    start_time = MPI_Wtime();
    collide();
    collide_time = MPI_Wtime();
    border_exchange();
    exchange_time = MPI_Wtime();
    stream();
    stream_time = MPI_Wtime();
    itertime = MPI_Wtime();

    collide_total += collide_time - start_time;
    exchange_total += exchange_time - collide_time;
    stream_total += stream_time - exchange_time;

    /* Dump velocity field for
     * visualization every snap_freq
     * steps */
    if (quiet == false && (iter % snap_freq) == 0) {
      // Ignore IO
      MPI_Pcontrol(0);
      mpi_store(iter / snap_freq);

      if (world_rank == 0) {
        if (!quiet)
          printf("Iteration %" INT_FMT " / %" INT_FMT " (%.f %%). Time last "
                 "%" INT_FMT "iters: %f \n",
                 iter,
                 max_iter,
                 100.0 * iter / (float)max_iter,
                 snap_freq,
                 itertime - last_snap);
      }
      last_snap = MPI_Wtime();
      MPI_Pcontrol(1);
    }
  }

  /* Store only last state if I/O is
   * disabled */
  if (quiet == true)
    mpi_store(max_iter / snap_freq);

  MPI_Pcontrol(0);
  LIKWID_MARKER_CLOSE;
  // -------------- PROFILING STOP
  // -----------------

  double red_collide, red_exchange, red_stream;
  MPI_Reduce(
    &collide_total, &red_collide, 1, MPI_DOUBLE, MPI_SUM, 0, comm_cart);
  MPI_Reduce(&stream_total, &red_stream, 1, MPI_DOUBLE, MPI_SUM, 0, comm_cart);
  MPI_Reduce(
    &exchange_total, &red_exchange, 1, MPI_DOUBLE, MPI_SUM, 0, comm_cart);

  if (world_rank == 0) {
    printf("TOTAL TIME: \t\t %lf\n", red_collide + red_exchange + red_stream);
    printf("\tCOLLIDE TIME: \t %lf\n", red_collide);
    printf("\tSTREAM TIME: \t %lf\n", red_stream);
    printf("\tEXCHANGE TIME: \t %lf\n", red_exchange);
    printf("\n");
    printf("\tCOLLIDE [ITER]: \t%lf\n", collide_total / max_iter);
    printf("\tSTREAM [ITER]:  \t%lf\n", stream_total / max_iter);
    printf("\tEXCHANGE [ITER]: \t%lf\n", exchange_total / max_iter);
  }

  mpi_write_perf(collide_total, stream_total, exchange_total, max_iter);

  /* Release the memory and halt */
  free(density[0]);
  free(density[1]);
  free(index_arr);
  free(velocity);
  free(map);
  free(map_final);
  free(abs_velocity);
  free(v_x);
  free(v_y);
  free(v_z);

#ifdef P_COORDS
  free(p_coords);
#endif
  free(rank);

  MPI_Finalize();

  exit(EXIT_SUCCESS);
}

void
border_exchange(void)
{
  // South
  //   ^
  // North
  MPI_Sendrecv(density[1],
               1,
               mpi_t_plane_N,
               NB_N,
               0,
               density[1],
               1,
               mpi_t_plane_S_ghost,
               NB_S,
               0,
               comm_cart,
               MPI_STATUS_IGNORE);

  // South
  //   v
  // North
  MPI_Sendrecv(density[1],
               1,
               mpi_t_plane_S,
               NB_S,
               0,
               density[1],
               1,
               mpi_t_plane_N_ghost,
               NB_N,
               0,
               comm_cart,
               MPI_STATUS_IGNORE);

  // East <- West
  MPI_Sendrecv(density[1],
               1,
               mpi_t_plane_W,
               NB_W,
               0,
               density[1],
               1,
               mpi_t_plane_E_ghost,
               NB_E,
               0,
               comm_cart,
               MPI_STATUS_IGNORE);

  // East -> West
  MPI_Sendrecv(density[1],
               1,
               mpi_t_plane_E,
               NB_E,
               0,
               density[1],
               1,
               mpi_t_plane_W_ghost,
               NB_W,
               0,
               comm_cart,
               MPI_STATUS_IGNORE);
}

/* Collision/relaxation operator:
 * redistribute the densities at each lattice point so that they
 * approach equilibrium.
 */
void
collide(void)
{

  LIKWID_MARKER_START("collide");
#pragma omp parallel for
#ifdef OCOLLIDE_LOOP_INDIRECT
  for (int_t n = 0; n < fluid_node_count; n++) {
    int_t k = p_coords[n].z;
    int_t i = p_coords[n].y;
    int_t j = p_coords[n].x;
    if (i < 0 || k < 0 || i > height - 1 || k > depth - 1)
      continue;
#else
  for (int_t k = 0; k < depth; k++)
    for (int_t i = 0; i < height; i++)
      for (int_t j = 0; j < width; j++) {
#endif
    int_t global_j = j /* +  cart_coords[1] * width */;
    int_t global_i = cart_coords[0] * height;
    real_t rho = 0.0, uc = 0.0;
    V(i, j, k, 0) = 0.0;
    V(i, j, k, 1) = 0.0;
    V(i, j, k, 2) = 0.0;

    /* Disregard solid points */
    if (MAP(i, j, k) == FLUID || MAP(i, j, k) == WALL) {

      /* Calculate the current velocity
       * field in the fluid */
      if (MAP(i, j, k) == FLUID) {
        for (int_t d = 0; d < D_COUNT; d++) {
          rho += D_now(i, j, k, d);
          V(i, j, k, 0) += c[d][0] * D_now(i, j, k, d);
          V(i, j, k, 1) += c[d][1] * D_now(i, j, k, d);
          V(i, j, k, 2) += c[d][2] * D_now(i, j, k, d);
        }
        V(i, j, k, 0) /= rho;
        V(i, j, k, 1) /= rho;
        V(i, j, k, 2) /= rho;
      }

      /* Calculate collision operator,
         inspired by "A Coupled Approach
         for Fluid Dynamic Problems
          Using the PDE Framework
         Peano", Neumann et al.
       */
      for (int_t d = 0; d < D_COUNT; d++) {
        real_t N_eq, delta_N;
        uc = c[d][0] * V(i, j, k, 0) + c[d][1] * V(i, j, k, 1) +
             c[d][2] * V(i, j, k, 2);
        N_eq =
          w[d] * rho *
          (1.0 + (uc / (c_s * c_s)) +
           ((uc * uc) / (2.0 * c_s * c_s * c_s * c_s)) -
           ((V(i, j, k, 1) * V(i, j, k, 1) + V(i, j, k, 0) * V(i, j, k, 0) +
             V(i, j, k, 2) * V(i, j, k, 2)) /
            (2 * c_s * c_s)));

        delta_N = lambda * (D_now(i, j, k, d) - N_eq);

        /* External force at j=1 */
        if (global_j == 1 && MAP(i, j, k) == FLUID)
          delta_N += w[d] * (c[d][0] * force[0] + c[d][1] * force[1] +
                             c[d][2] * force[2]); // 1e-3;

        switch (MAP(i, j, k)) {
          /* Redistribute fluid according to density/velocity */
          case FLUID:
            D_nxt(i, j, k, d) = D_now(i, j, k, d) + delta_N;
            break;
          /* Walls reflect incoming mass in opposite direction */
          case WALL:
            if (d != 0)
              D_nxt(i, j, k, bounce(d)) = D_now(i, j, k, d);
            break;
          /* No work to do on solid points */
          case SOLID:
            break;
        }
      }
    }
  }

  LIKWID_MARKER_STOP("collide");
}

void
stream(void)
{
  LIKWID_MARKER_START("stream");
#pragma omp parallel for
#ifdef OSTREAM_LOOP_INDIRECT
  for (int_t n = 0; n < fluid_node_count; n++) {
    int_t k = p_coords[n].z;
    int_t i = p_coords[n].y;
    int_t j = p_coords[n].x;
#else
  for (int_t k = -1; k < depth + 1; k++)
    for (int_t i = -1; i < height + 1; i++)
      for (int_t j = 0; j < width; j++) {
#endif
    for (int_t d = 0; d < D_COUNT; d++) {
      if (P_index(i, j, k) < 0)
        continue;

      int_t ni = (i + OFFSET(d, 0));
      int_t nj = (j + OFFSET(d, 1) + width) % width; // wrap around in x
      int_t nk = (k + OFFSET(d, 2));

      if (ni < 0 || nk < 0 || ni > height - 1 || nk > depth - 1)
        continue;

      if (P_index(ni, nj, nk) < 0)
        continue;

      /* Propagate present fluid density
       * to neighbors
       */
      D_now(ni, nj, nk, d) = D_nxt(i, j, k, d);
    }
  }
  LIKWID_MARKER_STOP("stream");
}

void
mpi_store(int_t iter)
{
  for (int_t k = 0; k < depth; k++) {
    for (int_t i = 0; i < height; i++) {
      for (int_t j = 0; j < width; j++) {

        v_x[k * width * height + i * width + j] = V(i, j, k, 0);
        v_y[k * width * height + i * width + j] = V(i, j, k, 1);
        v_z[k * width * height + i * width + j] = V(i, j, k, 2);

        V_abs(i, j, k) =
          sqrt(V(i, j, k, 0) * V(i, j, k, 0) + V(i, j, k, 1) * V(i, j, k, 1) +
               V(i, j, k, 2) * V(i, j, k, 2));
      }
    }
  }

  char filename[256];
  memset(filename, 0, 256);
  sprintf(filename, "data/d3q%d-%.5" INT_FMT ".nc", D_COUNT, iter);

  int err;
  err = nc_create_par(filename, NC_NETCDF4, comm_cart, MPI_INFO_NULL, &ncid);
  ERR err = nc_def_dim(ncid, "d1", g_depth, &dimid[0]);
  ERR err = nc_def_dim(ncid, "d2", g_height, &dimid[1]);
  ERR err = nc_def_dim(ncid, "d3", g_width, &dimid[2]);
  ERR err = nc_def_var(ncid, "map", NC_INT, 3, dimid, &var_map);
  ERR err = nc_def_var(ncid, "velocity", NC_DOUBLE, 3, dimid, &varid);
  ERR err = nc_def_var(ncid, "v_z", NC_DOUBLE, 3, dimid, &var_vz);
  ERR err = nc_def_var(ncid, "v_y", NC_DOUBLE, 3, dimid, &var_vy);
  ERR err = nc_def_var(ncid, "v_x", NC_DOUBLE, 3, dimid, &var_vx);
  ERR err = nc_def_var(ncid, "rank", NC_INT, 3, dimid, &var_rank);
  ERR err = nc_enddef(ncid);
  ERR

    err = nc_var_par_access(ncid, var_map, NC_COLLECTIVE);
  ERR err = nc_var_par_access(ncid, varid, NC_COLLECTIVE);
  ERR err = nc_var_par_access(ncid, var_vx, NC_COLLECTIVE);
  ERR err = nc_var_par_access(ncid, var_vy, NC_COLLECTIVE);
  ERR err = nc_var_par_access(ncid, var_vz, NC_COLLECTIVE);
  ERR err = nc_var_par_access(ncid, var_rank, NC_COLLECTIVE);
  ERR

    int global_i = cart_coords[0] * height;
  int global_j =
    /* cart_coords[1] * width */ 0;
  int global_k = cart_coords[1] * depth;
  size_t start[3] = { global_k, global_i, global_j };
  size_t count[3] = { depth, height, width };
  // ptrdiff_t stride[3] = { 1, 1, 1 };

  nc_put_vara_int(ncid, var_rank, start, count, rank);
  nc_put_vara_int(ncid, var_map, start, count, (int*)&MAP_I(0, 0, 0));
  nc_put_vara_double(ncid, varid, start, count, &V_abs(0, 0, 0));

  nc_put_vara_double(ncid, var_vy, start, count, v_y);
  nc_put_vara_double(ncid, var_vx, start, count, v_x);
  nc_put_vara_double(ncid, var_vz, start, count, v_z);

  nc_close(ncid);
}

/*
 * Find the bounding box of the domain.
 * Assumes that there is only extra padding in the large x, y and z values
 *
 * Will modify depth, width and height to the minimum.
 */
void
find_bounding_box(Binvox_content* voxel_map,
                  int_t* depth,
                  int_t* width,
                  int_t* height)
{
  int_t bound_width, bound_height, bound_depth;

  bound_depth = bound_width = bound_height = 0;

  for (int_t i = 0; i < voxel_map->height; i++)
    for (int_t j = 0; j < voxel_map->width; j++)
      for (int_t k = 0; k < voxel_map->depth; k++) {
        if (VOXEL_MAP(voxel_map, i, j, k)) {
          if (i > bound_height)
            bound_height = i;
          if (j > bound_width)
            bound_width = j;
          if (k > bound_depth)
            bound_depth = k;
        }
      }

  for (int_t i = 0; i < 1; i++) {
    bound_depth =
      bound_depth < voxel_map->depth ? bound_depth + 1 : voxel_map->depth;
    bound_width =
      bound_width < voxel_map->width ? bound_width + 1 : voxel_map->width;
    bound_height =
      bound_height < voxel_map->height ? bound_height + 1 : voxel_map->height;
  }

  *depth = bound_depth;
  *width = bound_width;
  *height = bound_height;
}

void
exchange_map_ghost_cells(void)
{

  MPI_Sendrecv(&MAP(0, 0, -1),
               1,
               row_dep_plane_INT,
               NB_N,
               0,
               &MAP(height, 0, -1),
               1,
               row_dep_plane_INT,
               NB_S,
               0,
               comm_cart,
               MPI_STATUS_IGNORE);
  MPI_Sendrecv(&MAP(height - 1, 0, -1),
               1,
               row_dep_plane_INT,
               NB_S,
               0,
               &MAP(-1, 0, -1),
               1,
               row_dep_plane_INT,
               NB_N,
               0,
               comm_cart,
               MPI_STATUS_IGNORE);

  MPI_Sendrecv(&MAP(-1, 0, 0),
               1,
               row_col_plane_INT,
               NB_W,
               0,
               &MAP(-1, 0, depth),
               1,
               row_col_plane_INT,
               NB_E,
               0,
               comm_cart,
               MPI_STATUS_IGNORE);
  MPI_Sendrecv(&MAP(-1, 0, depth - 1),
               1,
               row_col_plane_INT,
               NB_E,
               0,
               &MAP(-1, 0, -1),
               1,
               row_col_plane_INT,
               NB_W,
               0,
               comm_cart,
               MPI_STATUS_IGNORE);
}

void
initialize_premade_domain(Binvox_content* voxel_map)
{
  for (int_t k = -1; k < depth + 1; k++)
    for (int_t i = -1; i < height + 1; i++)
      for (int_t j = 0; j < width; j++) {
        MAP(i, j, k) = SOLID;
        P_index(i, j, k) = -1;
      }

  for (int_t k = 0; k < depth; k++) {
    for (int_t i = 0; i < height; i++) {
      for (int_t j = 0; j < width; j++) {
        int_t binvox_i = i + cart_coords[0] * height;
        int_t binvox_j = j;
        int_t binvox_k = k + cart_coords[1] * depth;

        bool oubG = binvox_i < 0 || binvox_k < 0 || binvox_i > g_height - 1 ||
                    binvox_k > g_depth - 1;

        if (!oubG && VOXEL_MAP(voxel_map, binvox_i, binvox_j, binvox_k))
          MAP(i, j, k) = FLUID;
        else
          MAP(i, j, k) = SOLID;

        P_index(i, j, k) = -1;
      }
    }
  }

  /* Check all solid points for fluid
   * neighbors, and classify them as
   * WALL if they have any.
   */
  for (int_t k = 0; k < depth; k++)
    for (int_t i = 0; i < height; i++) {
      for (int_t j = 0; j < width; j++) {
        if (MAP(i, j, k) == SOLID) {
          for (int_t d = 0; d < D_COUNT; d++) {
            int_t ni = (i + OFFSET(d, 0));
            int_t nj = (j + OFFSET(d, 1));
            int_t nk = (k + OFFSET(d, 2));
            if (ni < 0 || nj < 0 || nk < 0 || ni > height - 1 ||
                nj > width - 1 || nk > depth - 1)
              continue;
            if (MAP(ni, nj, nk) == FLUID) {
              MAP(i, j, k) = WALL;
              break;
            }
          }
        }
      }
    }

  exchange_map_ghost_cells();

  for (int_t k = -1; k < depth + 1; k++)
    for (int_t i = -1; i < height + 1; i++)
      for (int_t j = 0; j < width; j++) {
        if (MAP(i, j, k) != SOLID) {
#ifdef P_COORDS
          p_coords[fluid_node_count].x = j;
          p_coords[fluid_node_count].y = i;
          p_coords[fluid_node_count].z = k;
#endif
          P_index(i, j, k) = fluid_node_count++;
        }
      }
}

/*
 * Box for poisseuille flow
 */
void
create_box()
{
  // Create top/bottom and side walls. No walls in the x direction
#pragma omp parallel for
  for (int_t k = 0; k < depth; k++)
    for (int_t i = 0; i < height; i++)
      for (int_t j = 0; j < width; j++) {
        int_t global_i = (i + cart_coords[0] * height) % g_height;
        int_t global_k = (k + cart_coords[1] * depth) % g_depth;

        if (global_i == 0 || global_i == g_depth - 1 || global_k == 0 ||
            global_k == g_depth - 1) {
          MAP(i, j, k) = SOLID;
        } else {
          MAP(i, j, k) = FLUID;
        }
      }
}

/*
 * Create a cylinder obstruction, extruded from x-y plane
 */
void
create_cylinder(int_t y_center, int_t x_center, int_t radius_sq)
{
#pragma omp parallel for
  for (int_t k = 0; k < depth; k++)
    for (int_t i = 0; i < height; i++)
      for (int_t j = 0; j < width; j++) {
        int_t global_i = (i + cart_coords[0] * height) % g_height;
        int_t global_j = j;

        real_t dist_from_center =
          (global_i - y_center) * (global_i - y_center) +
          (global_j - x_center) * (global_j - x_center);

        if (dist_from_center < radius_sq + 1) {
          MAP(i, j, k) = SOLID;
        }
      }
}

/*
 * Configure the domain as a channel with a solid circular obstruction
 * positioned at 1/4 of its length.
 */
void
create_pipe()
{
  int_t center[DIMS] = { (g_height / 2), (g_width / 4), (g_depth / 2) };

  // Assumes square/identical z and y
  if (pipe_radius <= 0.0)
    pipe_radius = ((real_t)g_height / 2) - 5;

  /* Solid pipe */
  for (int_t k = 0; k < depth; k++) {
    for (int_t i = 0; i < height; i++) {
      for (int_t j = 0; j < width; j++) {
        int_t global_i = (i + cart_coords[0] * height) % g_height;
        int_t global_j = j;
        int_t global_k = (k + cart_coords[1] * depth) % g_depth;

        real_t dist_from_center =
          sqrt((global_i - center[0]) * (global_i - center[0]) +
               (global_j - center[1]) * (global_j - center[1]) +
               (global_k - center[2]) * (global_k - center[2]));

        bool is_outside_pipe =
          (global_i - center[0]) * (global_i - center[0]) +
            (global_k - center[2]) * (global_k - center[2]) >=
          pipe_radius * pipe_radius;

        // Ball in the middle
        if (dist_from_center < radius)
          /*MAP(i, j, k) = FLUID;*/
          MAP(i, j, k) = SOLID;
        //  Walls around z/2 and y/2 in
        //  x-direction
        else if (is_outside_pipe)
          MAP(i, j, k) = SOLID;
        else
          MAP(i, j, k) = FLUID;
      }
    }
  }
}

void
create_binvox(char* file)
{
  Binvox_content* contents = malloc(sizeof(Binvox_content));
  contents->data = malloc(depth * width * height);
  contents->depth = depth;
  contents->height = height;
  contents->width = width;

  for (int_t i = 0; i < height; i++)
    for (int_t j = 0; j < width; j++)
      for (int_t k = 0; k < depth; k++) {
        int_t index = j * depth * height + i * depth + k;
        if (MAP(i, j, k) == SOLID)
          contents->data[index] = (char)0;
        else
          contents->data[index] = (char)1;
      }

  write_binvox(file, contents);
  free(contents->data);
  free(contents);
}

void
create_plane(int_t z)
{
  for (int_t k = 0; k < depth; k++)
    for (int_t i = 0; i < height; i++)
      for (int_t j = 0; j < width; j++) {
        int_t global_k = (k + cart_coords[1] * depth) % g_depth;

        if (global_k == z) {
          MAP(i, j, k) = SOLID;
        }
      }
}

#define CYL_AREA_PART 0.005

void
initialize_domain(void)
{

  // Fill initial values
  for (int_t k = -1; k < depth + 1; k++)
    for (int_t i = -1; i < height + 1; i++)
      for (int_t j = 0; j < width; j++) {
        MAP(i, j, k) = SOLID;
        P_index(i, j, k) = -1;
      }

  if (domain_settings.type == POISSEUILLE) {
    create_box();

    int_t solid_count = 0;
    int_t target_solid = (100 - domain_settings.node_usage);
    int i = 1;
    double cyl_area = CYL_AREA_PART;

    while (solid_count <
           (width * height * depth) * ((real_t)target_solid / 100)) {
      printf("Runing create. solid count: %ld. Target: %lf\n",
             solid_count,
             (width * height * depth) * ((real_t)target_solid / 100));
      int_t cyl_radius_sq = (cyl_area * g_depth * g_width) / 3.14;
      int cyl_count = ((double)target_solid / i) / (cyl_area * 100) + 1;
      int rand = random();
      if (i < 5)
        i++;
      if (i > 4) {
        for (int i = 0; i < 5; i++)
          create_plane(((depth / 10) * i + rand % depth) % depth);
      }
      printf("Creating %d cylinders\n", cyl_count);
      for (int i = 0; i < cyl_count; i++) {

        int_t cyl_radius_max = sqrt(cyl_radius_sq) * 1.1;
        int_t y_center = random() % (g_height);
        int_t x_center =
          (i * cyl_radius_max * 2 + (rand % cyl_radius_max)) % (g_width);
        create_cylinder(y_center, x_center, cyl_radius_sq);
      }
      solid_count = 0;
      for (int_t k = 0; k < depth; k++)
        for (int_t i = 0; i < height; i++)
          for (int_t j = 0; j < width; j++) {
            if (MAP(i, j, k) == SOLID) {
              int nogo = 0;
              for (int_t d = 0; d < D_COUNT; d++) {
                int_t ni = (i + OFFSET(d, 0));
                int_t nj = (j + OFFSET(d, 1));
                int_t nk = (k + OFFSET(d, 2));
                if (ni < 0 || nj < 0 || nk < 0 || ni > height - 1 ||
                    nj > width - 1 || nk > depth - 1)
                  continue;
                if (MAP(ni, nj, nk) == FLUID) {

                  nogo = 1;
                  break;
                }
              }
              if (nogo == 0)
                solid_count++;
            }
          }
      printf("Created %ld points\n", solid_count);
    }

    if (store_binvox)

      /*MPI_Allgather(MAP(0, 0, 0), );*/

      if (cart_rank == 0)
        create_binvox(binvox_path);
  } else if (domain_settings.type == PIPE) {
    create_box();
    int_t center[DIMS - 1] = { (g_height / 2), (g_width / 4) };
    create_cylinder(center[0], center[1], radius * radius);

  } else {
    create_pipe();
  }

  /* Check all solid points for fluid
   * neighbors, and classify them as
   * WALL if they have any.
   */
  for (int_t k = 0; k < depth; k++)
    for (int_t i = 0; i < height; i++) {
      for (int_t j = 0; j < width; j++) {

        rank[k * width * height + i * width + j] = cart_rank;

        if (MAP(i, j, k) == SOLID) {
          for (int_t d = 0; d < D_COUNT; d++) {
            int_t ni = (i + OFFSET(d, 0));
            int_t nj = (j + OFFSET(d, 1));
            int_t nk = (k + OFFSET(d, 2));
            if (ni < 0 || nj < 0 || nk < 0 || ni > height - 1 ||
                nj > width - 1 || nk > depth - 1)
              continue;
            if (MAP(ni, nj, nk) == FLUID) {
              MAP(i, j, k) = WALL;
              break;
            }
          }
        }
      }
    }

  exchange_map_ghost_cells();

  for (int_t k = -1; k < depth + 1; k++)
    for (int_t i = -1; i < height + 1; i++)
      for (int_t j = 0; j < width; j++) {
        if (MAP(i, j, k) != SOLID) {
#ifdef P_COORDS
          p_coords[fluid_node_count].x = j;
          p_coords[fluid_node_count].y = i;
          p_coords[fluid_node_count].z = k;
#endif
          P_index(i, j, k) = fluid_node_count++;
        }
      }
}

void
mpi_write_perf(double collide_total,
               double stream_total,
               double exchange_total,
               int iterations)
{
  MPI_File fh;
  int char_count = 300;

  char filename[256] = "lbm-perf.csv";
  if (strcmp(perf_file, "") != 0) {
    strcpy(filename, perf_file);
  }

  MPI_File_open(
    comm_cart, filename, MPI_MODE_CREATE | MPI_MODE_RDWR, MPI_INFO_NULL, &fh);

  char header[char_count];
  int header_len = 0;
  if (cart_rank == 0) {
    printf("[lbm-perf] Storing performance characteristics in %s\n", filename);
    sprintf(header,
            "rank,fluid_nodes,solid_nodes,porosity,collide_iter,stream_iter,"
            "exchange_iter,collide_total,stream_total,exchange_total\n");
    header_len = strlen(header);
  }

  MPI_File_write_ordered(fh, &header, header_len, MPI_CHAR, MPI_STATUS_IGNORE);

  char rank_line[char_count];

  int_t solid_node_count = height * width * depth - fluid_node_count;
  double porosity = (double)fluid_node_count / (height * width * depth);
  double collide_iter = collide_total / iterations;
  double stream_iter = stream_total / iterations;
  double exchange_iter = exchange_total / iterations;

  sprintf(rank_line,
          "%d,%ld,%ld,%lf,%lf,%lf,%lf,%lf,%lf,%lf\n",
          cart_rank,
          fluid_node_count,
          solid_node_count,
          porosity,
          collide_iter,
          stream_iter,
          exchange_iter,
          collide_total,
          stream_total,
          exchange_total);

  MPI_File_write_ordered(
    fh, &rank_line, strlen(rank_line), MPI_CHAR, MPI_STATUS_IGNORE);

  MPI_File_close(&fh);
}

/* Command line options and usage information */
static const char* usage =
  "    -h                    : Usage information\n"
  "    -q                    : Suppress input/output\n"
  "    -s <integer>          : Save state every <integer> steps (default 100)\n"
  "    -H <integer>          : Set domain height (default 40)\n"
  "    -W <integer>          : Set domain width  (default 80)\n"
  "    -D <integer>          : Set domain depth  (default 40)\n"
  "    -R <integer>          : Set obstruction radius (default height/20)\n"
  "    -F <real>             : Set horizontal force magnitude (default 1e-2)\n"
  "    -I <integer>          : Set iteration count (default 40001)\n"
  "    -t <integer>          : Set number of threads per CPU (OpenMP)\n"
  "    -u <integer>          : Set percentage of nodes to be fluid\n"
  "    -d                    : Stop execution and print PID for debugging\n"
  "    -b <path>             : Use a 3d voxel file for environment\n"
  "\n"
  "    --store-binvox <path> : Store resulting domain as a binvox file\n"
  "    --domain-type <type>  : choose the domain type to generate\n"
  "    --cart-dimsx <int>    : Dimensions of MPI cart in x direction\n"
  "    --cart-dimsy <int>    : Dimensions of MPI cart in y direction\n";

void
options(int argc, char** argv)
{
  static struct option long_options[] = {
    { "store-binvox", required_argument, 0, 0 },
    { "domain-type", required_argument, 0, 0 },
    { "cart-dimsy", required_argument, 0, 0 },
    { "cart-dimsx", required_argument, 0, 0 },
    { "perf-file", required_argument, 0, 0 },
  };
  int option_idx = 0;
  int o;
  while (
    (o = getopt_long(
       argc, argv, "qhds:H:W:D:R:F:I:t:b:u:", long_options, &option_idx)) !=
    -1) {
    switch (o) {
      case 0:
        switch (option_idx) {
          case 0:
            strcpy(binvox_path, optarg);
            store_binvox = true;
            break;
          case 1:
            if (strcmp(optarg, "pipe") == 0) {
              domain_settings.type = PIPE;
            }
          case 2:
            cart_dims_opt[0] = strtol(optarg, NULL, 10);
            break;
          case 3:
            cart_dims_opt[1] = strtol(optarg, NULL, 10);
            break;
          case 4:
            strcpy(perf_file, optarg);
            break;

          default:
            break;
        }
        break;
      case 'q':
        quiet = true;
        break;
      case 's':
        snap_freq = strtol(optarg, NULL, 10);
        break;
      case 'H':
        g_height = strtol(optarg, NULL, 10);
        break;
      case 'W':
        g_width = strtol(optarg, NULL, 10);
        break;
      case 'D':
        g_depth = strtol(optarg, NULL, 10);
        break;
      case 'R':
        radius = (real_t)strtol(optarg, NULL, 10);
        break;
      case 'F':
        force[1] = (real_t)strtod(optarg, NULL);
        break;
      case 'I':
        max_iter = strtol(optarg, NULL, 10);
        break;
      case 't':
        threads = strtol(optarg, NULL, 10);
        break;
      case 'u':
        domain_settings.node_usage = strtol(optarg, NULL, 10);
        break;
      case 'd':
        debug = true;
        break;
      case 'b':
        strcpy(binvox_path, optarg);
        load_binvox_data = true;
        break;
      case 'h':
        fprintf(stderr, "%s\n%s", argv[0], usage);
        exit(EXIT_FAILURE);
        break;
      default:
        exit(EXIT_FAILURE);
        break;
    }
  }
}
