import argparse
import csv
import re
import os
import statistics
import math
from typing import Dict, List

parser = argparse.ArgumentParser()
parser.add_argument("--path", help="Path to data folder", type=str)
parser.add_argument("--key", help="Key to sort by", type=str)
parser.add_argument("--reverse", help="Reverse sort order", action="store_true")
parser.add_argument("-p", help="Choose processor", type=str)


def read_data(file: str) -> List[Dict[str, str]]:
    data = []

    with open(file) as f:
        reader = csv.DictReader(f)
        for row in reader:
            data.append(row | {"filename": os.path.basename(file)})

    return data


def process_perf(path: str):
    data_files = list(os.listdir(path))

    data = [read_data(os.path.join(path, d)) for d in data_files]

    res = []
    versions = {}

    for d in data:
        collide_sum = 0.0
        stream_sum = 0.0
        exchange_sum = 0.0
        porosity_sum = 0.0
        for row in d:
            collide_sum += float(row["collide_iter"])
            stream_sum += float(row["stream_iter"])
            exchange_sum += float(row["exchange_iter"])
            porosity_sum += float(row["porosity"])

        collide_avg = collide_sum / len(d)
        stream_avg = stream_sum / len(d)
        exchange_avg = exchange_sum / len(d)
        porosity_avg = porosity_sum / len(d)

        p = re.compile(".*-([a-zA-Z0-9]+)-([a-zA-Z0-9]+[-_].*)\.csv")
        re_res = p.findall(d[0]["filename"])
        version = re_res[0][0]
        model = re_res[0][1]

        if "RCA" in model:
            model = "rca"
        else:
            model = re.findall(r"(p\d+)-\d+", model)[0]

        item = {
            "name": version,
            "porosity": round(porosity_avg, 2),
            f"collide_{model}": collide_avg,
            f"stream_{model}": stream_avg,
            f"exchange_{model}": exchange_avg,
        }

        if version in versions:
            versions[version].update(item)
        else:
            versions.update({version: item})

    ordering = ["base", "IA", "IALO", "IALO1D"]
    for version in ordering:
        res.append(versions[version])

    return res


def process(path: str):
    data_files = list(os.listdir(path))

    data = [read_data(os.path.join(path, d)) for d in data_files]
    fix_p = False
    if path == "data/density-base-400":
        fix_p = True

    res = []

    for d in data:
        collide_sum = 0.0
        stream_sum = 0.0
        exchange_sum = 0.0
        porosity_sum = 0.0
        for row in d:
            collide_sum += float(row["collide_iter"])
            stream_sum += float(row["stream_iter"])
            exchange_sum += float(row["exchange_iter"])
            porosity_sum += float(row["porosity"])

        collide_avg = collide_sum / len(d)
        stream_avg = stream_sum / len(d)
        exchange_avg = exchange_sum / len(d)
        porosity_avg = porosity_sum / len(d)

        total = collide_sum + stream_sum + exchange_sum
        total_err = statistics.stdev(
            [
                float(a["collide_iter"])
                + float(a["stream_iter"])
                + float(a["exchange_iter"])
                for a in d
            ]
        )

        # Forgot to fix baseline...
        if porosity_avg == 0.0 or fix_p:
            p = re.compile("perf-(\d+)\.csv")
            porosity_avg = float(p.findall(d[0]["filename"])[0]) / 100

        res.append(
            {
                "porosity": round(porosity_avg, 2),
                "collide": collide_avg,
                "stream": stream_avg,
                "exchange": exchange_avg,
                "total": total / len(d),
                "total_err": total_err,
            }
        )

    return res


def process_ranks(path: str):
    data_files = list(os.listdir(path))

    data = [read_data(os.path.join(path, d)) for d in data_files]

    res = []

    for d in data:
        collide_sum = 0.0
        stream_sum = 0.0
        exchange_sum = 0.0
        porosity_sum = 0.0
        for row in d:
            collide_sum += float(row["collide_iter"])
            stream_sum += float(row["stream_iter"])
            exchange_sum += float(row["exchange_iter"])
            porosity_sum += float(row["porosity"])

        collide_avg = collide_sum / len(d)
        stream_avg = stream_sum / len(d)
        exchange_avg = exchange_sum / len(d)
        porosity_avg = porosity_sum / len(d)
        total = collide_sum + stream_sum + exchange_sum
        total_err = statistics.stdev(
            [
                float(a["collide_iter"])
                + float(a["stream_iter"])
                + float(a["exchange_iter"])
                for a in d
            ]
        )

        res.append(
            {
                "ranks": len(d),
                "porosity": round(porosity_avg, 4),
                "collide": collide_avg,
                "stream": stream_avg,
                "exchange": exchange_avg,
                "total": total / len(d),
                "total_err": total_err,
            }
        )

    return res


def get_porosity(path: str):
    data = read_data(path)

    total_fluid = 0
    total_solid = 0
    max_p = 0
    min_p = math.inf
    for row in data:
        total_fluid += int(row["fluid_nodes"])
        total_solid += int(row["solid_nodes"])

        max_p = max(max_p, float(row["porosity"]))
        min_p = min(min_p, float(row["porosity"]))

    total_p = total_fluid / (total_fluid + total_solid)
    print(f"Total: {round(total_p, 3)}        ({total_p})")
    print(f"min  : {round(min_p, 3)}          ({min_p})")
    print(f"max  : {round(max_p, 3)}          ({max_p})")


def write_data(path, data, sort_key, *, reverse=False):
    res = data
    if sort_key is not None:
        res = sorted(data, key=lambda a: a[sort_key], reverse=reverse)

    with open(f"{path}-comb.csv", "w") as f:
        writer = csv.DictWriter(f, fieldnames=res[0].keys())
        writer.writeheader()
        writer.writerows(res)


def main():
    args = parser.parse_args()
    path = args.path
    reverse = False
    if args.reverse:
        reverse = True

    res = None
    match args.p:
        case "ranks":
            res = process_ranks(path)
        case "density":
            res = process(path)
        case "general":
            res = process_perf(path)
        case "porosity":
            get_porosity(path)
            return
        case _:
            print("Please specify processor with -p")
            exit(1)

    write_data(path, res, args.key, reverse=reverse)


if __name__ == "__main__":
    main()
