#define _GNU_SOURCE ;
#include "read_binvox.h"
#include <errno.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

int
read_binvox(char* filename, Binvox_content* contents)
{

  FILE* fd;

  if (access(filename, R_OK) != 0) {
    fprintf(stderr,
            "[binvox load]: Could not open file for reading: %s - %s\n",
            filename,
            strerror(errno));
    return 1;
  }

  fd = fopen(filename, "r");

  char* linebuf = NULL;
  size_t linesize = 0;

  char* saveptr;
  char* next;
  int depth, height, width = -1;

  bool done = false;
  while (!done) {
    getline(&linebuf, &linesize, fd);

    next = strtok_r(linebuf, " \n", &saveptr);
    if (strcmp(next, "data") == 0)
      break;

    if (strcmp(next, "dim") == 0) {

      next = strtok_r(NULL, " ", &saveptr);
      depth = strtol(next, NULL, 10);

      next = strtok_r(NULL, " ", &saveptr);
      width = strtol(next, NULL, 10);

      next = strtok_r(NULL, " ", &saveptr);
      height = strtol(next, NULL, 10);
    }
  }

  free(linebuf);

  if (width == -1) {

    fprintf(stderr, "[binvox load]: ERR could not find dimensions\n");
    return 1;
  }

  contents->width = width;
  contents->height = height;
  contents->depth = depth;

  int size = width * height * depth;
  char* voxels = malloc(size);
  contents->data = voxels;

  int64_t index = 0, end = 0, nr_voxels = 0;
  int value, count;

  printf("[binvox load]: Loading %d points from file\n", size);

  while (end < size) {
    // Binvox data contains of pairs of bytes, the first being the voxel value
    // (0/1) and the second the number of repeating voxels with that value.
    value = fgetc(fd);
    if (value == EOF) {
      fprintf(stderr, "[binvox load]: ERR Reached EOF before done parsing\n");
      goto error;
    }
    count = (u_int8_t)fgetc(fd);

    end = index + count;
    if (value)
      nr_voxels += count;

    // Fill count number of voxels with byte value
    for (int i = index; i < end; i++)
      voxels[i] = value;

    index = end;
  }

  fclose(fd);

  printf("read %ld voxels from %s\n", nr_voxels, filename);

  return 0;

error:
  free(voxels);
  return 1;
}

int
write_binvox(char* filename, Binvox_content* contents)
{
  FILE* fd;

  /*
  if (access(filename, W_OK) != 0) {
    fprintf(stderr,
            "[binvox load]: Could not open file for writing: %s - %s\n",
            filename,
            strerror(errno));
    return 1;
  }
  */

  fd = fopen(filename, "wb");
  fprintf(fd,
          "#binvox 1\n"
          "dim %d %d %d\n"
          "data\n",
          contents->depth,
          contents->width,
          contents->height);

  char current_value;
  current_value = contents->data[0];
  u_int8_t current_count = 1;
  int size = contents->depth * contents->width * contents->height;
  for (int i = 1; i < size; i++) {
    if (contents->data[i] != current_value || current_count == 255) {
      fwrite(&current_value, 1, 1, fd);
      fwrite(&current_count, 1, 1, fd);

      current_value = contents->data[i];
      current_count = 1;
    } else {
      current_count++;
    }
  }

  fwrite(&current_value, 1, 1, fd);
  fwrite(&current_count, 1, 1, fd);

  printf("[binvox] Wrote %d voxels to %s\n", size, filename);

  fclose(fd);

  return 0;
}

#ifdef BINVOX_MAIN

int
main(int argc, char** argv)
{
  if (argc < 2) {
    printf("Usage: ./program [filename]");
  }

  Binvox_content* data = malloc(sizeof(Binvox_content));
  if (read_binvox(argv[1], data) == 0) {
    free(data->data);
  }
  free(data);
}

#endif
